import collections
import itertools
import logging
import typing

from odd.artifact import Confidence, Issue, Location, XMLRecord
from odd.plugin import Plugin
from odd.utils import split_external_id
from odd_bunch.xml_utils import drop_comments, get_view_arch

_LOG = logging.getLogger(__name__)
VIEW_TAGS = frozenset(
    [
        "tree",
        "form",
        "graph",
        "pivot",
        "kanban",
        "calendar",
        "search",
        # These views don't have `<field>`:
        # "gantt", "diagram", "dashboard", "cohort".
    ]
)
# These tags can have `<field>` inside,
FIELD_TAGS = frozenset(["templates", "button"])


def find_fields(el, path):
    result = []
    for child in el.iterchildren():
        child_tag = child.tag
        if child.tag == "field":
            field_name = child.attrib["name"]
            result.append(("/".join((*path, field_name)), child.sourceline))
            child_tag = f"field[{field_name}]"

        new_path = path
        if child.tag in VIEW_TAGS or child.tag in FIELD_TAGS or child.tag != child_tag:
            new_path = (*path, child_tag)
        result.extend(find_fields(child, new_path))
    return result


def has_single_item(iterable: typing.Iterable) -> bool:
    return len(tuple(itertools.islice(iterable, 0, 2))) == 1


class DuplicateViewFields(Plugin):
    _handles = {"xml_record"}
    _emits = {"issue"}

    def on_xml_record(self, xml_record: XMLRecord):
        record = xml_record.node
        if record.attrib["model"] != "ir.ui.view":
            return
        view_xml_id = record.get("id")

        # Skip inherited views.
        inherit_id = record.xpath("./field[@name='inherit_id']")
        if inherit_id and inherit_id[0].attrib.get("ref"):
            return

        # Skip `arch` override in an extending addon.
        if view_xml_id:
            addon_name, _ = split_external_id(view_xml_id)
            if addon_name and addon_name != xml_record.addon.name:
                return

        arch = get_view_arch(record)
        if arch is None:
            _LOG.warning(
                "`ir.ui.view` record has no `arch` field " "in file: %s at line %d",
                xml_record.path,
                record.sourceline,
            )
            return

        if not has_single_item(drop_comments(arch.iterchildren())):
            _LOG.warning(
                "Unexpected number of children in `ir.ui.view` "
                "`arch` in file: %s at line %d",
                xml_record.path,
                arch.sourceline,
            )
            return

        fields: typing.Dict[
            typing.Tuple[str], typing.List[int]
        ] = collections.defaultdict(list)
        for path, line_no in find_fields(arch, ()):
            fields[path].append(line_no)

        for path, line_nos in fields.items():
            if len(line_nos) > 1:
                field_name = path.split("/")[-1]
                yield Issue(
                    "duplicate_view_field",
                    f'"{view_xml_id}" `ir.ui.view` has duplicate field '
                    f'"{field_name}"'
                    if view_xml_id
                    else (f'`ir.ui.view` has duplicate field "{field_name}"'),
                    xml_record.addon,
                    locations=[
                        Location(xml_record.path, [line_no]) for line_no in line_nos
                    ],
                    confidence=Confidence.LOW,
                    categories=["correctness"],
                )
