from odd.artifact import Issue, Location, XMLRecord
from odd.plugin import Plugin
from odd.utils import split_external_id

from odd_bunch.xml_utils import collect_fields


class IrCronStateCode(Plugin):
    _handles = {"xml_record"}
    _emits = {"issue"}

    def on_xml_record(self, xml_record: XMLRecord):
        model, external_id = (xml_record.node.get("model"), xml_record.node.get("id"))
        if xml_record.addon.version < 11 or model != "ir.cron":
            return

        is_extension = False
        if external_id:
            addon, _ = split_external_id(external_id)
            is_extension = addon and addon != xml_record.addon.name

        record = xml_record.node
        fields = {
            name: (line_no, text) for name, line_no, _, text in collect_fields(record)
        }

        # Do nothing if the record is an extension and both `code` and `state` fields
        # are not being set, since either can be set on the original record.
        if is_extension and ("code" not in fields or "state" not in fields):
            return

        code_line_no, code = fields.get("code", (None, None))
        state_line_no, state = fields.get("state", (None, None))
        if code and not state:
            yield Issue(
                "incorrect_cron_record",
                f'`ir.cron` record "{external_id}" has `code` set, '
                f"but no `state` set - code will not be executed when the "
                f"cron job runs",
                xml_record.addon,
                locations=[Location(xml_record.path, [code_line_no])],
                categories=["correctness"],
            )
        elif code and state != "code":
            yield Issue(
                "incorrect_cron_record",
                f'`ir.cron` record "{external_id}" has `code` set, '
                f'but `state` is not "code" - code will not be executed '
                f"when the cron job runs",
                xml_record.addon,
                locations=[
                    Location(xml_record.path, [code_line_no]),
                    Location(xml_record.path, [state_line_no]),
                ],
                categories=["correctness"],
            )
        elif not code and state == "code":
            yield Issue(
                "incorrect_cron_record",
                f'`ir.cron` record "{external_id}" has `state` set '
                f'to "code", but `code` is not set',
                xml_record.addon,
                locations=[Location(xml_record.path, [state_line_no])],
                categories=["correctness"],
            )
