import collections

from odd.artifact import Issue, Location, PythonModule
from odd.ast_utils import get_string_node_value, is_string_node, iter_node_type
from odd.const import MANIFEST_FILENAMES, SUPPORTED_VERSIONS
from odd.plugin import Plugin
from odd.utils import column_index_1, expand_version_list

KEY_VERSION_MAP = expand_version_list(
    {
        ">=8": {
            "name",
            "version",
            "description",
            "author",
            "website",
            "license",
            "category",
            "depends",
            "data",
            "demo",
            "auto_install",
            "external_dependencies",
            "application",
            "css",
            "images",
            "installable",
            "maintainer",
            "summary",
            "qweb",
            "sequence",
            "bootstrap",
            "icon",
            # Hooks.
            "pre_init_hook",
            "post_init_hook",
            "uninstall_hook",
            "post_load",
            # Not sure about the purpose of this.
            "web",
            # TODO: Check if this makes sense in >= v12.
            "test",
            # Deprecated.
            "active",
            "init_xml",
            "update_xml",
            "demo_xml",
            # App store.
            "support",
            "price",
            "currency",
            "iap",
            "live_test_url",
            # OCA.
            "maintainers",
            "development_status",
        },
        ">=11": {"excludes"},
    },
    *SUPPORTED_VERSIONS,
    result_cls=set,
)
# Keys which contain lists that may contain duplicated values.
LIST_KEYS = frozenset(
    (
        "css",
        "data",
        "demo",
        "demo_xml",
        "depends",
        "excludes",
        "images",
        "init",
        "init_xml",
        "qweb",
        "test",
        "update_xml",
    )
)


class ManifestKeys(Plugin):
    _handles = {"python_module"}
    _emits = {"issue"}

    def _check_active(self, addon):
        if "active" in addon.manifest:
            yield "active", {
                "slug": "deprecated_manifest_key",
                "description": '"active" manifest key was renamed to "auto_install"',
                "categories": ["correctness", "deprecated"],
            }

    def _check_deprecated_xml(self, addon):
        for key in ("init_xml", "update_xml", "demo_xml"):
            if key in addon.manifest:
                if key == "init_xml":
                    files = addon.manifest.get(key) or []
                    if not any(fn.endswith(".xml") for fn in files):
                        continue
                correct_key = "demo" if key == "demo_xml" else "data"
                yield key, {
                    "slug": "deprecated_manifest_key",
                    "description": (
                        f'"{key}" manifest key was deprecated '
                        f'in favor of "{correct_key}"'
                    ),
                    "categories": ["correctness", "deprecated"],
                }

    def _check_unknown_keys(self, addon):
        unknown_keys = addon.manifest.keys() - KEY_VERSION_MAP[addon.version]
        for key in unknown_keys:
            yield key, {
                "slug": "unknown_manifest_key",
                "description": f'Unknown manifest key "{key}"',
                "categories": ["correctness"],
            }

    def _check_duplicate_values(self, addon):
        for key in LIST_KEYS:
            for value, count in collections.Counter(
                addon.manifest.get(key) or []
            ).most_common():
                if count < 2:
                    break
                yield key, {
                    "slug": "duplicate_manifest_value",
                    "description": (
                        f'"{key}" manifest key has a duplicated value: {value}'
                    ),
                    "categories": ["correctness"],
                }

    def _get_key_locations(self, module):
        key_locations = collections.defaultdict(list)
        for node in iter_node_type(module, "dictorsetmaker"):
            for child in node.children[::4]:
                if is_string_node(child):
                    key_locations[get_string_node_value(child)].append(
                        column_index_1(child.start_pos)
                    )
            break
        return key_locations

    def on_python_module(self, python_module: PythonModule):
        if python_module.path.name not in MANIFEST_FILENAMES:
            return
        key_locations = None
        for check in ("active", "deprecated_xml", "unknown_keys", "duplicate_values"):
            for key, issue in getattr(self, f"_check_{check}")(python_module.addon):
                if key_locations is None:
                    key_locations = self._get_key_locations(python_module.node)
                yield Issue(
                    **{
                        "locations": [Location(python_module.path, key_locations[key])],
                        "addon": python_module.addon,
                        **issue,
                    }
                )
