import logging
import pathlib
import typing

from odd.artifact import Addon, Issue, Location, Severity, XMLDocument, XMLRecord
from odd.plugin import Plugin
from odd.utils import split_external_id

from odd_bunch.utils import odoo_reference_url

_LOG = logging.getLogger(__name__)
ACTION_MODELS = frozenset(
    (
        "ir.actions.act_window",
        "ir.actions.act_window.view",
        "ir.actions.act_url",
        "ir.actions.server",
        "ir.actions.client",
    )
)


class RedundantMenuitemName(Plugin):
    """
    Checks for uses of the `<menuitem>` shortcut where the `name` attribute
    value matches the `name` of the action record (linked via the `action`
    attribute (if any)). In such cases the `name` on the `<menuitem>` can be
    omitted.
    """

    _handles = {"xml_document", "xml_record"}
    _emits = {"issue"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._actions: typing.Dict[
            str, typing.Tuple[pathlib.Path, int, typing.Optional[str]]
        ] = {}
        self._menuitems: typing.List[
            typing.Tuple[pathlib.Path, int, typing.Optional[str], str, str]
        ] = []

    def _extract_xml_act_window(self, addon, path, tree):
        # <act_window> shortcut.
        for act_window in tree.xpath("//act_window"):
            external_id = act_window.get("id")
            if external_id:
                _, record_id = split_external_id(external_id)
                self._actions[record_id] = (
                    path,
                    act_window.sourceline,
                    act_window.get("name"),
                )

    def _extract_xml_menuitem(self, addon, path, tree):
        # <menuitem> shortcut.
        for menuitem in tree.xpath("//menuitem"):
            external_id = menuitem.get("id")
            name = menuitem.get("name")
            action_external_id = menuitem.get("action")
            if not (name and action_external_id):
                continue

            # We currently do not have a way to get action names from other
            # addons, so skip if the referenced addon is not this one.
            action_addon_name, action_record_id = split_external_id(action_external_id)
            if action_addon_name and action_addon_name != addon.name:
                continue

            self._menuitems.append(
                (path, menuitem.sourceline, external_id, name, action_record_id)
            )

    def on_xml_document(self, xml_document: XMLDocument):
        for op in ("act_window", "menuitem"):
            getattr(self, f"_extract_xml_{op}")(
                xml_document.addon, xml_document.path, xml_document.node
            )
        yield from ()

    def on_xml_record(self, xml_record: XMLRecord):
        model = xml_record.node.get("model")
        external_id = xml_record.node.get("id")
        if model not in ACTION_MODELS or not external_id:
            return

        name, line_no = "", xml_record.node.sourceline
        for field in xml_record.node.iterchildren(tag="field"):
            if field.get("name") == "name" and field.text:
                name, line_no = field.text, field.sourceline
                break

        _, record_id = split_external_id(external_id)
        self._actions[external_id] = xml_record.path, line_no, name

        yield from ()

    def on_after(self, addon: Addon):
        for (
            menu_path,
            menu_line_no,
            menu_external_id,
            menu_name,
            action_record_id,
        ) in self._menuitems:
            action_values = self._actions.get(action_record_id)
            if action_values is None:
                _LOG.warning(
                    'Action "%s", referenced in `<menuitem>` "%s" in file: %s at line: '
                    '%d was not discovered by "redundant_menuitem_name" plugin, '
                    "is missing or is declared on an unexpected model",
                    action_record_id,
                    menu_external_id,
                    menu_path,
                    menu_line_no,
                )
                continue
            action_path, action_line_no, action_name = action_values

            if menu_name == action_name:
                yield Issue(
                    "redundant_menuitem_name",
                    f'`name` attribute defined on the "{menu_external_id}" '
                    f"`ir.ui.menu` record can be omitted (the value will be taken "
                    f'from the linked action record "{action_record_id}")',
                    addon,
                    locations=[
                        Location(menu_path, [menu_line_no]),
                        Location(action_path, [action_line_no]),
                    ],
                    severity=Severity.INFO,
                    categories=["redundancy"],
                    sources=[
                        odoo_reference_url(addon.version, "data", fragment="menuitem")
                    ],
                )
