import importlib.resources
import logging

import lxml.etree as ET  # nosec

from odd.artifact import Confidence, Issue, Location, Severity, XMLDocument
from odd.const import SUPPORTED_VERSIONS
from odd.plugin import Plugin
from odd_bunch.xml_utils import get_root

_LOG = logging.getLogger(__name__)


VERSION_RNG_MAP = {}
for version in SUPPORTED_VERSIONS:
    with importlib.resources.path(
        f"odd_bunch.data.relaxng.v{version:d}", f"import_xml.rng"
    ) as path:
        _LOG.debug("Loading RelaxNG schema from file: %s", path)
        VERSION_RNG_MAP[version] = ET.RelaxNG(get_root(path))


class RelaxNG(Plugin):
    _handles = {"xml_document"}
    _emits = {"issue"}

    def on_xml_document(self, xml_document: XMLDocument):
        relaxng = VERSION_RNG_MAP[xml_document.addon.version]

        try:
            relaxng.assert_(xml_document.node)
        except AssertionError:
            last_error = relaxng.error_log.last_error
            yield Issue(
                "relaxng_error",
                f"XML file does not match Odoo RelaxNG schema: {last_error.message}",
                xml_document.addon,
                locations=[
                    Location(
                        xml_document.path, [(last_error.line, last_error.column + 1)]
                    )
                ],
                # The addon will fail to install/update if the XML file fails
                # to validate.
                severity=Severity.MAJOR,
                # We use the same RelaxNG schema which is used by Odoo.
                confidence=Confidence.HIGH,
                categories=["correctness"],
            )
